"""
Tests for ``tahoe invite``.
"""

import json
import os
from os.path import join
from typing import Optional, Sequence

from twisted.internet import defer
from twisted.trial import unittest

from ...client import read_config
from ...scripts import runner
from ...util.jsonbytes import dumps_bytes
from ..common_util import run_cli
from ..no_network import GridTestMixin
from .common import CLITestMixin
from .wormholetesting import MemoryWormholeServer, memory_server


class Join(GridTestMixin, CLITestMixin, unittest.TestCase):

    @defer.inlineCallbacks
    def setUp(self):
        self.basedir = self.mktemp()
        yield super(Join, self).setUp()
        yield self.set_up_grid(oneshare=True)

    @defer.inlineCallbacks
    def test_create_node_join(self):
        """
        successfully join after an invite
        """
        node_dir = self.mktemp()
        server = MemoryWormholeServer()
        options = runner.Options()
        options.wormhole = server
        reactor = object()

        wormhole = server.create(
            "tahoe-lafs.org/invite",
            "ws://wormhole.tahoe-lafs.org:4000/v1",
            reactor,
        )
        code = yield wormhole.get_code()
        messages = [
            {u"abilities": {u"server-v1": {}}},
            {
                u"shares-needed": 1,
                u"shares-happy": 1,
                u"shares-total": 1,
                u"nickname": u"somethinghopefullyunique",
                u"introducer": u"pb://foo",
            },
        ]
        for msg in messages:
            wormhole.send_message(dumps_bytes(msg))

        rc, out, err = yield run_cli(
            "create-client",
            "--join", code,
            node_dir,
            options=options,
        )

        self.assertEqual(0, rc)

        config = read_config(node_dir, u"")
        self.assertIn(
            "pb://foo",
            set(
                furl
                for (furl, cache)
                in config.get_introducer_configuration().values()
            ),
        )

        with open(join(node_dir, 'tahoe.cfg'), 'r') as f:
            config = f.read()
        self.assertIn(u"somethinghopefullyunique", config)

    @defer.inlineCallbacks
    def test_create_node_illegal_option(self):
        """
        Server sends JSON with unknown/illegal key
        """
        node_dir = self.mktemp()
        server = MemoryWormholeServer()
        options = runner.Options()
        options.wormhole = server
        reactor = object()

        wormhole = server.create(
            "tahoe-lafs.org/invite",
            "ws://wormhole.tahoe-lafs.org:4000/v1",
            reactor,
        )
        code = yield wormhole.get_code()
        messages = [
            {u"abilities": {u"server-v1": {}}},
            {
                u"shares-needed": 1,
                u"shares-happy": 1,
                u"shares-total": 1,
                u"nickname": u"somethinghopefullyunique",
                u"introducer": u"pb://foo",
                u"something-else": u"not allowed",
            },
        ]
        for msg in messages:
            wormhole.send_message(dumps_bytes(msg))

        rc, out, err = yield run_cli(
            "create-client",
            "--join", code,
            node_dir,
            options=options,
        )

        # should still succeed -- just ignores the not-whitelisted
        # "something-else" option
        self.assertEqual(0, rc)


class Invite(GridTestMixin, CLITestMixin, unittest.TestCase):

    @defer.inlineCallbacks
    def setUp(self):
        self.basedir = self.mktemp()
        yield super(Invite, self).setUp()
        yield self.set_up_grid(oneshare=True)
        intro_dir = os.path.join(self.basedir, "introducer")
        yield run_cli(
            "create-introducer",
            "--listen", "none",
            intro_dir,
        )

    async def _invite_success(self, extra_args=(), tahoe_config=None):
        # type: (Sequence[bytes], Optional[bytes]) -> defer.Deferred
        """
        Exercise an expected-success case of ``tahoe invite``.

        :param extra_args: Positional arguments to pass to ``tahoe invite``
            before the nickname.

        :param tahoe_config: If given, bytes to write to the node's
            ``tahoe.cfg`` before running ``tahoe invite.
        """
        intro_dir = os.path.join(self.basedir, "introducer")
        # we've never run the introducer, so it hasn't created
        # introducer.furl yet
        priv_dir = join(intro_dir, "private")
        with open(join(priv_dir, "introducer.furl"), "w") as fobj_intro:
            fobj_intro.write("pb://fooblam\n")
        if tahoe_config is not None:
            assert isinstance(tahoe_config, bytes)
            with open(join(intro_dir, "tahoe.cfg"), "wb") as fobj_cfg:
                fobj_cfg.write(tahoe_config)

        wormhole_server, helper = memory_server()
        options = runner.Options()
        options.wormhole = wormhole_server
        reactor = object()

        async def server():
            # Run the server side of the invitation process using the CLI.
            rc, out, err = await run_cli(
                "-d", intro_dir,
                "invite",
                *tuple(extra_args) + ("foo",),
                options=options,
            )

        async def client():
            # Run the client side of the invitation by manually pumping a
            # message through the wormhole.

            # First, wait for the server to create the wormhole at all.
            wormhole = await helper.wait_for_wormhole(
                "tahoe-lafs.org/invite",
                "ws://wormhole.tahoe-lafs.org:4000/v1",
            )
            # Then read out its code and open the other side of the wormhole.
            code = await wormhole.when_code()
            other_end = wormhole_server.create(
                "tahoe-lafs.org/invite",
                "ws://wormhole.tahoe-lafs.org:4000/v1",
                reactor,
            )
            other_end.set_code(code)

            # Send a proper client abilities message.
            other_end.send_message(dumps_bytes({u"abilities": {u"client-v1": {}}}))

            # Check the server's messages.  First, it should announce its
            # abilities correctly.
            server_abilities = json.loads(await other_end.when_received())
            self.assertEqual(
                server_abilities,
                {
                    "abilities":
                    {
                        "server-v1": {}
                    },
                },
            )

            # Second, it should have an invitation with a nickname and
            # introducer furl.
            invite = json.loads(await other_end.when_received())
            self.assertEqual(
                invite["nickname"], "foo",
            )
            self.assertEqual(
                invite["introducer"], "pb://fooblam",
            )
            return invite

        invite, _ = await defer.gatherResults(map(
            defer.Deferred.fromCoroutine,
            [client(), server()],
        ))
        return invite


    @defer.inlineCallbacks
    def test_invite_success(self):
        """
        successfully send an invite
        """
        invite = yield defer.Deferred.fromCoroutine(self._invite_success((
            "--shares-needed", "1",
            "--shares-happy", "2",
            "--shares-total", "3",
        )))
        self.assertEqual(
            invite["shares-needed"], "1",
        )
        self.assertEqual(
            invite["shares-happy"], "2",
        )
        self.assertEqual(
            invite["shares-total"], "3",
        )

    @defer.inlineCallbacks
    def test_invite_success_read_share_config(self):
        """
        If ``--shares-{needed,happy,total}`` are not given on the command line
        then the invitation is generated using the configured values.
        """
        invite = yield defer.Deferred.fromCoroutine(self._invite_success(tahoe_config=b"""
[client]
shares.needed = 2
shares.happy = 4
shares.total = 6
"""))
        self.assertEqual(
            invite["shares-needed"], "2",
        )
        self.assertEqual(
            invite["shares-happy"], "4",
        )
        self.assertEqual(
            invite["shares-total"], "6",
        )


    @defer.inlineCallbacks
    def test_invite_no_furl(self):
        """
        Invites must include the Introducer FURL
        """
        intro_dir = os.path.join(self.basedir, "introducer")

        options = runner.Options()
        options.wormhole = None

        rc, out, err = yield run_cli(
            "-d", intro_dir,
            "invite",
            "--shares-needed", "1",
            "--shares-happy", "1",
            "--shares-total", "1",
            "foo",
            options=options,
        )
        self.assertNotEqual(rc, 0)
        self.assertIn(u"Can't find introducer FURL", out + err)

    @defer.inlineCallbacks
    def test_invite_wrong_client_abilities(self):
        """
        Send unknown client version
        """
        intro_dir = os.path.join(self.basedir, "introducer")
        # we've never run the introducer, so it hasn't created
        # introducer.furl yet
        priv_dir = join(intro_dir, "private")
        with open(join(priv_dir, "introducer.furl"), "w") as f:
            f.write("pb://fooblam\n")

        wormhole_server, helper = memory_server()
        options = runner.Options()
        options.wormhole = wormhole_server
        reactor = object()

        async def server():
            rc, out, err = await run_cli(
                "-d", intro_dir,
                "invite",
                "--shares-needed", "1",
                "--shares-happy", "1",
                "--shares-total", "1",
                "foo",
                options=options,
            )
            self.assertNotEqual(rc, 0)
            self.assertIn(u"No 'client-v1' in abilities", out + err)

        async def client():
            # Run the client side of the invitation by manually pumping a
            # message through the wormhole.

            # First, wait for the server to create the wormhole at all.
            wormhole = await helper.wait_for_wormhole(
                "tahoe-lafs.org/invite",
                "ws://wormhole.tahoe-lafs.org:4000/v1",
            )
            # Then read out its code and open the other side of the wormhole.
            code = await wormhole.when_code()
            other_end = wormhole_server.create(
                "tahoe-lafs.org/invite",
                "ws://wormhole.tahoe-lafs.org:4000/v1",
                reactor,
            )
            other_end.set_code(code)

            # Send some surprising client abilities.
            other_end.send_message(dumps_bytes({u"abilities": {u"client-v9000": {}}}))

        yield defer.gatherResults(map(
            defer.Deferred.fromCoroutine,
            [client(), server()],
        ))


    @defer.inlineCallbacks
    def test_invite_no_client_abilities(self):
        """
        Client doesn't send any client abilities at all
        """
        intro_dir = os.path.join(self.basedir, "introducer")
        # we've never run the introducer, so it hasn't created
        # introducer.furl yet
        priv_dir = join(intro_dir, "private")
        with open(join(priv_dir, "introducer.furl"), "w") as f:
            f.write("pb://fooblam\n")

        wormhole_server, helper = memory_server()
        options = runner.Options()
        options.wormhole = wormhole_server
        reactor = object()

        async def server():
            # Run the server side of the invitation process using the CLI.
            rc, out, err = await run_cli(
                "-d", intro_dir,
                "invite",
                "--shares-needed", "1",
                "--shares-happy", "1",
                "--shares-total", "1",
                "foo",
                options=options,
            )
            self.assertNotEqual(rc, 0)
            self.assertIn(u"No 'abilities' from client", out + err)

        async def client():
            # Run the client side of the invitation by manually pumping a
            # message through the wormhole.

            # First, wait for the server to create the wormhole at all.
            wormhole = await helper.wait_for_wormhole(
                "tahoe-lafs.org/invite",
                "ws://wormhole.tahoe-lafs.org:4000/v1",
            )
            # Then read out its code and open the other side of the wormhole.
            code = await wormhole.when_code()
            other_end = wormhole_server.create(
                "tahoe-lafs.org/invite",
                "ws://wormhole.tahoe-lafs.org:4000/v1",
                reactor,
            )
            other_end.set_code(code)

            # Send a no-abilities message through to the server.
            other_end.send_message(dumps_bytes({}))

        yield defer.gatherResults(map(
            defer.Deferred.fromCoroutine,
            [client(), server()],
        ))


    @defer.inlineCallbacks
    def test_invite_wrong_server_abilities(self):
        """
        Server sends unknown version
        """
        intro_dir = os.path.join(self.basedir, "introducer")
        # we've never run the introducer, so it hasn't created
        # introducer.furl yet
        priv_dir = join(intro_dir, "private")
        with open(join(priv_dir, "introducer.furl"), "w") as f:
            f.write("pb://fooblam\n")

        wormhole_server = MemoryWormholeServer()
        options = runner.Options()
        options.wormhole = wormhole_server
        reactor = object()

        wormhole = wormhole_server.create(
            "tahoe-lafs.org/invite",
            "ws://wormhole.tahoe-lafs.org:4000/v1",
            reactor,
        )
        code = yield wormhole.get_code()
        messages = [
            {u"abilities": {u"server-v9000": {}}},
            {
                "shares-needed": "1",
                "shares-total": "1",
                "shares-happy": "1",
                "nickname": "foo",
                "introducer": "pb://fooblam",
            },
        ]
        for msg in messages:
            wormhole.send_message(dumps_bytes(msg))

        rc, out, err = yield run_cli(
            "create-client",
            "--join", code,
            "foo",
            options=options,
        )
        self.assertNotEqual(rc, 0)
        self.assertIn("Expected 'server-v1' in server abilities", out + err)

    @defer.inlineCallbacks
    def test_invite_no_server_abilities(self):
        """
        Server sends unknown version
        """
        intro_dir = os.path.join(self.basedir, "introducer")
        # we've never run the introducer, so it hasn't created
        # introducer.furl yet
        priv_dir = join(intro_dir, "private")
        with open(join(priv_dir, "introducer.furl"), "w") as f:
            f.write("pb://fooblam\n")

        server = MemoryWormholeServer()
        options = runner.Options()
        options.wormhole = server
        reactor = object()

        wormhole = server.create(
            "tahoe-lafs.org/invite",
            "ws://wormhole.tahoe-lafs.org:4000/v1",
            reactor,
        )
        code = yield wormhole.get_code()
        messages = [
            {},
            {
                "shares-needed": "1",
                "shares-total": "1",
                "shares-happy": "1",
                "nickname": "bar",
                "introducer": "pb://fooblam",
            },
        ]
        for msg in messages:
            wormhole.send_message(dumps_bytes(msg))

        rc, out, err = yield run_cli(
            "create-client",
            "--join", code,
            "bar",
            options=options,
        )
        self.assertNotEqual(rc, 0)
        self.assertIn("Expected 'abilities' in server introduction", out + err)

    @defer.inlineCallbacks
    def test_invite_no_nick(self):
        """
        Should still work if server sends no nickname
        """
        intro_dir = os.path.join(self.basedir, "introducer")

        options = runner.Options()
        options.wormhole = None

        rc, out, err = yield run_cli(
            "-d", intro_dir,
            "invite",
            "--shares-needed", "1",
            "--shares-happy", "1",
            "--shares-total", "1",
            options=options,
        )
        self.assertTrue(rc)
        self.assertIn(u"Provide a single argument", out + err)
